#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of svgis.
# https://github.com/fitnr/svgis

# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://www.opensource.org/licenses/GNU General Public License v3 (GPLv3)-license
# Copyright (c) 2015, Neil Freeman <contact@fakeisthenewreal.org>

from . import test_convert
from . import test_draw
from . import test_projection
from . import test_svg



"""Draw geodata layers into svg"""
# -*- coding: utf-8 -*-
from os import path
from collections import Iterable
import fiona
import fiona.transform
import svgwrite
import fionautil.coords
from . import projection
from . import draw
from . import svg
from . import scale
from . import convert
from . import osm

STYLE = ('polyline, line, rect, path, polygon, .polygon {'
         ' fill: none;'
         ' stroke: #000;'
         ' stroke-width: 1px;'
         ' stroke-linejoin: round;'
         '}')


def _draw_feature(geom, properties=None, **kwargs):
    '''Draw a single feature given a geometry object and properties object'''
    attribs = {}
    properties = properties or {}

    if kwargs.get('classes'):
        try:
            attribs['class'] = ' '.join(str(properties.get(c, '')).replace(' ', '_') for c in kwargs.pop('classes'))
        except TypeError:
            pass

    if kwargs.get('id_field'):
        try:
            attribs['id'] = str(properties.get(kwargs.pop('id_field'))).replace(' ', '_')
        except AttributeError:
            pass

    ps = draw.geometry(geom, **kwargs)

    if len(ps) > 1:
        target = svgwrite.container.Group()
        for p in ps:
            target.add(p)
    else:
        target = ps[0]

    target.attribs.update(attribs)
    return target


class SVGIS(object):

    """Draw geodata files to SVG"""

    in_crs = dict()
    bbox = dict()

    def __init__(self, files, bounds=None, out_crs=None, **kwargs):
        if isinstance(files, basestring):
            self.files = [files]
        elif isinstance(files, Iterable):
            self.files = files
        else:
            raise ValueError("'files' must be a file name or list of file names")

        if bounds:
            self.bbox['bbox'] = bounds

        self.mbr = (None, None, None, None)

        self.out_crs = out_crs

        self.use_proj = kwargs.pop('use_proj', None)

        self.scalar = kwargs.pop('scalar', 1)

        self.style = kwargs.pop('style', STYLE)

        self.padding = kwargs.pop('padding', 0)

    def compose_file(self, filename, scalar, **kwargs):
        '''Draw file to svg
        filename -- a fiona-readable file
        mbr -- a tuple containing (minx, maxx, miny, maxy) in the layer's coordinate system. 'None' values are OK
        '''
        if filename[-4:] == path.extsep + 'osm':
            return self._compose_osm(filename, scalar, **kwargs)
        else:
            return self._compose_fiona(filename, scalar, **kwargs)

    def _compose_osm(self, filename, scalar, **kwargs):
        kwargs.pop('id_field', None)
        kwargs.pop('classes', None)

        osm_root = osm.get_root(filename)
        bounds = osm.bounds(osm_root)

        if not self.out_crs:
            self.out_crs = projection.choosecrs(osm.CRS, bounds, use_proj=self.use_proj)

        self.mbr = convert.updatebounds(self.mbr, projection.project_mbr(osm.CRS, self.out_crs, *bounds))

        group = osm.draw(osm_root, scalar, out_crs=self.out_crs, **kwargs)
        group.attribs['id'] = filename

        return group

    def _compose_fiona(self, filename, scalar, **kwargs):
        '''Draw file to svg
        filename -- a fiona-readable file
        mbr -- a tuple containing (minx, maxx, miny, maxy) in the layer's coordinate system. 'None' values are OK
        '''
        with fiona.open(filename, "r") as layer:
            group = svgwrite.container.Group(id=layer.name)

            self.in_crs[layer.name] = layer.crs

            if not self.out_crs:
                # Determine projection transformation:
                # either use something passed in, a non latlong layer projection,
                # the local UTM, or customize local TM
                self.out_crs = projection.choosecrs(layer.crs, layer.bounds, use_proj=self.use_proj)

            self.mbr = convert.updatebounds(self.mbr, projection.project_mbr(layer.crs, self.out_crs, *layer.bounds))

            if 'classes' in kwargs:
                kwargs['classes'] = [c for c in kwargs['classes'] if c in layer.schema['properties']]

            if 'id_field' in kwargs:
                if kwargs['id_field'] not in layer.schema['properties'].keys():
                    del kwargs['id_field']

            if self.out_crs != layer.crs:
                reproject = lambda geom: fiona.transform.transform_geom(layer.crs, self.out_crs, geom)
            else:
                reproject = lambda f: f

            for _, f in layer.items(**self.bbox):
                geom = scale.geometry(reproject(f['geometry']), scalar)
                target = _draw_feature(geom, f['properties'], **kwargs)
                group.add(target)

        return group

    def compose(self, style=None, scalar=None, **kwargs):
        '''Draw files to svg.
        scalar -- factor by which to scale the data.
        style -- CSS
        '''
        scalar = scalar or self.scalar
        style = style or self.style

        viewbox = kwargs.pop('viewbox', None)

        container = svgwrite.container.Group(transform='scale(1, -1)', fill_rule='evenodd')
        container.translate(self.padding, -self.padding)

        with fiona.drivers():
            for filename in self.files:
                container.add(self.compose_file(filename, scalar, **kwargs))

        w, h, x0, y1 = self.dims(scalar)

        if viewbox:
            drawing = svg.create((w, h), [container], style=style)
            drawing.viewbox(x0, -y1, w, h)

        else:
            container.translate(-x0, -y1)
            drawing = svg.create((w, h), [container], style=style)

        return drawing

    def dims(self, scalar):
        '''Calculate the width, height, origin X and max Y of the document'''
        mbr_ring = convert.mbr_to_bounds(*self.mbr)
        boundary = scale.scale(mbr_ring, scalar)

        x0, y0, x1, y1 = fionautil.coords.bounds(list(boundary))

        w = x1 - x0 + (self.padding * 2)
        h = y1 - y0 + (self.padding * 2)

        return w, h, x0, y1



import unittest
from svgis import projection


class ProjectionTestCase(unittest.TestCase):

    def testUtm(self):
        assert projection.utm_proj4(-21, 42) == '+proj=utm +zone=27 +north +datum=WGS84 +units=m +no_defs'

        assert projection.utm_proj4(-21, -42) == '+proj=utm +zone=27 +south +datum=WGS84 +units=m +no_defs'

    def testLocalTm(self):
        self.assertEqual(projection.tm_proj4(0, 0, 0), ('+proj=lcc +lon_0=0 +lat_1=0 +lat_2=0 +lat_0=0'
                                                        '+x_0=0 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0'
                                                        '+units=m +no_defs'))



import unittest
from svgis import svgis
import svgwrite
from pkg_resources import resource_filename

class SvgisTestCase(unittest.TestCase):

    def setUp(self):
        self.file = resource_filename('svgis', 'test_data/cb_2014_us_nation_20m.shp')

    def testSvgisCreate(self):
        svgis_obj = svgis.SVGIS(self.file)

        self.assertEqual(svgis_obj.files, [self.file])
        assert svgis_obj.mbr == (None,) * 4
        assert svgis_obj.out_crs == None
        assert svgis_obj.style == svgis.STYLE

        svgis_obj2 = svgis.SVGIS([self.file])
        assert svgis_obj2.files == [self.file]

        with self.assertRaises(ValueError):
            svgis.SVGIS(12)

    def testSvgisCompose(self):
        svgis_obj = svgis.SVGIS(self.file)
        composed = svgis_obj.compose()
        assert type(composed) == svgwrite.drawing.Drawing



"""Draw a geometries elements as SVG"""
# -*- coding: utf-8 -*-
from __future__ import division, print_function
from itertools import chain
try:
    import numpy as np
except ImportError:
    pass
import svgwrite
import fionautil.measure


def linestring(coordinates, precision=3, **kwargs):
    try:
        np.round(coordinates, precision)
    except NameError:
        coordinates = [(round(x, precision), round(y, precision)) for x, y in coordinates]

    return svgwrite.shapes.Polyline(coordinates, **kwargs)


def multilinestring(coordinates, **kwargs):
    return [linestring(coords, **kwargs) for coords in coordinates]


def lines(geom, **kwargs):
    if geom['type'] == 'LineString':
        return [linestring(geom['coordinates'], **kwargs)]

    elif geom['type'] == 'MultiLineString':
        return multilinestring(geom['coordinates'], **kwargs)


def path(coordinates, **kwargs):
    return svgwrite.path.Path(['M'] + coordinates, **kwargs)


def polygons(geom, **kwargs):
    '''Draw polygon(s) in a feature. transform is a function to operate on coords.
    Draws first ring clockwise, and subsequent ones counter-clockwise
    '''
    if geom['type'] == 'Polygon':
        return [polygon(geom['coordinates'], **kwargs)]

    elif geom['type'] == 'MultiPolygon':
        return multipolygon(geom['coordinates'], **kwargs)


def polygon(coordinates, precision=3, **kwargs):
    '''Draw an svg polygon based on coordinates.'''
    try:
        coordinates = np.round(np.array(coordinates), precision)

        if coordinates.shape[-1] > 2:
            shape = list(coordinates.shape)
            coordinates.resize(shape[0:-1] + [2])

    except NameError:
        coordinates = [[(round(x, precision), round(y, precision)) for x, y in ring] for ring in coordinates]

    if len(coordinates) == 1:
        try:
            return svgwrite.shapes.Polygon(coordinates[0], **kwargs)

        except ValueError:
            raise

    # This is trickier because drawing holes in SVG.
    # We go clockwise on the first ring, then counterclockwise
    if fionautil.measure.counterclockwise(coordinates[0]):
        coordinates[0] = coordinates[0][::-1]

    pth = path(list(coordinates[0]) + ['z'], fill_rule='evenodd', class_='polygon', **kwargs)

    for ring in coordinates[1:]:
        # make all interior run the counter-clockwise
        if fionautil.measure.clockwise(ring):
            ring = ring[::-1]

        pth.push(['M'] + list(ring) + ['z'])

    return pth


def multipolygon(coordinates, **kwargs):
    return [polygon(coords, **kwargs) for coords in coordinates]


def points(geom, **kwargs):
    kwargs['r'] = kwargs.get('r', 1)

    if geom['type'] == 'Point':
        return [point(geom['coordinates'], **kwargs)]

    elif geom['type'] == 'MultiPoint':
        return points(geom['coordinates'], **kwargs)


def point(coordinates, precision=3, **kwargs):
    try:
        x, y = coordinates.pop()
    except (AttributeError, TypeError):
        x, y = coordinates

    x, y = round(x, precision), round(y, precision)

    return svgwrite.shapes.Circle(center=(x, y), **kwargs)


def multipoint(coordinates, **kwargs):
    return [point((x, y), **kwargs) for x, y in coordinates]


def geometry(geom, **kwargs):
    '''Draw a geometry'''

    if geom['type'] == 'Point':
        return points(geom, **kwargs)

    elif geom['type'] in 'MultiPoint':
        return multipoint(geom, **kwargs)

    elif geom['type'] in ('LineString', 'MultiLineString'):
        return lines(geom, **kwargs)

    elif geom['type'] in ('Polygon', 'MultiPolygon'):
        return polygons(geom, **kwargs)

    else:
        raise ValueError("Can't draw features like this: {}".format(geom['type']))


def feature(feat, **kwargs):
    '''Draw a feature'''
    return geometry(feat['geometry'], **kwargs)


def group(features):
    '''Return a group with features drawn into it'''
    g = svgwrite.container.Group(fill_rule="evenodd")

    elems = [feature(f) for f in features]

    for e in chain(*elems):
        g.add(e)

    return g



from __future__ import division
import math


def rect(length, angle):
    '''polar to cartesian coordinates'''
    return length * math.cos(angle), length * math.sin(angle)


def updatebounds(old, new):
    '''
    Extend old with any more distant values from newpoints.
    Also replace any missing min/max points in old with values from new
    '''
    bounds = []

    # python3 gives TypeError when using None in min/max
    # This contraption avoids that problem.
    # List comp below replaces Nones in bounds with real values in new or old
    for n, m in zip(new[:2], old[:2]):
        try:
            bounds.append(min(n, m))
        except TypeError:
            bounds.append(None)

    for n, m in zip(new[2:], old[2:]):
        try:
            bounds.append(max(n, m))
        except TypeError:
            bounds.append(None)

    if any(not v for v in bounds):
        bounds = tuple((a or b or c) for a, b, c in zip(bounds, new, old))

    return bounds


def _frange(a, b, count=None):
    """Yield <count> points between two floats"""
    count = count or 10
    jump = (b - a) / 10

    while a < b:
        yield a
        a += jump


def mbr_to_bounds(minx, miny, maxx, maxy):
    """Convert min, max points to a boundary ring"""

    xs, ys = list(_frange(minx, maxx)), list(_frange(miny, maxy))

    left_top = [(minx, y) for y in ys] + [(x, maxy) for x in xs][1:]

    ys.reverse()
    xs.reverse()

    return left_top + [(maxx, y) for y in ys] + [(x, miny) for x in xs]



from __future__ import unicode_literals
import unittest
import re
from pkg_resources import resource_filename
from io import StringIO
from svgis import svg
from xml.dom import minidom
import svgwrite

class SvgTestCase(unittest.TestCase):

    def setUp(self):
        self.file = resource_filename('svgis', 'test_data/test.svg')
        self.newstyle = 'stroke {color:red;}'

    def test_add_style(self):
        new = svg.add_style(self.file, self.newstyle)
        css = minidom.parseString(new).getElementsByTagName('defs').item(0).getElementsByTagName('style').item(0)
        assert self.newstyle in css.toxml()

    def test_add_style_missing_def(self):
        with open(self.file) as f:
            replaced_svg = re.sub(r'<defs>.+?</defs>', '', f.read())

        io_svg = StringIO(replaced_svg)

        new = svg.add_style(io_svg, self.newstyle)
        css = minidom.parseString(new).getElementsByTagName('defs').item(0).getElementsByTagName('style').item(0)
        assert self.newstyle in css.toxml()

    def test_rescale(self):
        new = svg.rescale(self.file, 100)

        g = minidom.parseString(new).getElementsByTagName('g')[0]
        assert 'scale(100)' in g.attributes.get('transform').value

    def test_create(self):
        s = svg.create((100, 100), [])
        assert type(s) == svgwrite.drawing.Drawing

        s = svg.create(100, [])
        assert type(s) == svgwrite.drawing.Drawing

        s = svg.create(100, [], style=self.newstyle)
        assert type(s) == svgwrite.drawing.Drawing
        assert len(s.defs.elements) > 0
        assert self.newstyle in s.defs.elements[0].tostring()



'''Edit SVGs'''
# -*- coding: utf-8 -*-
import sys
import svgwrite.container
from collections import Sequence
from xml.dom import minidom
import fionautil.coords


def rescale(svgfile, factor):
    svg = minidom.parse(svgfile)

    scalar = 'scale({})'.format(factor)

    gs = svg.getElementsByTagName('g')[0]

    transform = gs.attributes.get('transform')

    if transform:
        transform.value = transform.value + ' ' + scalar
    else:
        gs.setAttribute('transform', scalar)

    return svg.toxml()


def dims(boundary, padding=0):
    '''Return width and height based on an boundary ring and an optional padding'''
    x0, y0, x1, y1 = fionautil.coords.bounds(boundary)

    w = x1 - x0 + (padding * 2)
    h = y1 - y0 + (padding * 2)

    return w, h


def add_style(svgfile, newstyle, replace=False):
    '''Add to the CSS style in an SVG file.
    svgfile -- Path to an SVG file
    newstyle -- CSS string
    replace -- (boolean) If true, replace the existing CSS with newstyle (default: False)
    '''
    svg = minidom.parse(svgfile)

    defs = svg.getElementsByTagName('defs').item(0)

    if not defs:
        defs = svg.createElement('defs')

        if not hasattr(svg, 'tagName'):
            elem = svg.getElementsByTagName('svg').item(0)
            elem.insertBefore(defs, elem.firstChild)

        else:
            svg.insertBefore(defs, svg.firstChild)

    if defs.getElementsByTagName('style'):
        style = defs.getElementsByTagName('style').item(0)

        if replace:
            style.firstChild.replaceWholeText(newstyle)
        else:
            style.firstChild.nodeValue += ' ' + newstyle

    else:
        style = svg.createElement('style')
        css = svg.createTextNode(newstyle)
        style.appendChild(css)
        defs.appendChild(style)

    return svg.toxml()


def set_group(members=None, scale=None, translate=None, **kwargs):
    '''Create a group with the given scale and translation'''

    groupargs = {
        'fill': 'none',
        'stroke': 'black',
    }
    groupargs.update(kwargs)
    g = svgwrite.container.Group(**groupargs)

    members = members or []

    for m in members:
        g.add(m)

    if scale:
        g.scale(*scale)

    if translate:
        g.translate(*translate)

    return g


def save(filename, diameter, groups, profile=None):
    dwg = create(diameter, groups, profile)
    dwg.filename = filename
    dwg.save()


def create(size, groups, profile=None, style=None):
    profile = profile or 'full'

    if not isinstance(size, Sequence):
        size = (int(size), int(size))

    dwg = svgwrite.Drawing(profile=profile, size=size)

    if style:
        dwg.defs.add(dwg.style(style))

    for g in groups:
        dwg.add(g)

    return dwg


def write(filename, drawing):

    try:
        if filename == '-':
            f = sys.stdout

        else:
            f = open(filename, 'w')

        f.write(drawing.tostring())

    except IOError:
        raise

    finally:
        try:
            if f != sys.stdout:
                f.close()
        except UnboundLocalError:
            pass



#!/usr/bin/python
from __future__ import print_function, division
import argparse
from . import svg
from .import projection
from .svgis import SVGIS
import sys


def _echo(content, output):
    if hasattr(output, 'write'):
        output.write(content)
    else:
        open(output, 'w').write(content)


def _scale(layer, output, scale, **_):
    '''Rescale an SVG by a factor'''
    result = svg.rescale(layer, scale)
    _echo(result, output)


def _style(layer, output, style, replace=None, **_):
    '''Add a CSS string to an SVG's style attribute'''
    result = svg.add_style(layer, style, replace=replace)
    _echo(result, output)


def _draw(layers, output, bounds=None, scale=1, padding=0, **kwargs):
    '''Draw a geodata layer to a simple SVG'''
    scalar = (1 / scale) if scale else 1
    out_crs = None

    use_proj = None

    if kwargs.get('epsg'):
        out_crs = 'EPSG:' + kwargs.pop('epsg')
    elif kwargs.get('proj4'):
        out_crs = kwargs.pop('proj4')
    else:
        use_proj = kwargs.pop('use_proj', None)

    # get rid of pesky kwargs
    for x in ('use_proj', 'epsg', 'proj4'):
        kwargs.pop(x, None)

    # Try to read style file
    if kwargs.get('style'):
        try:
            kwargs['style'] = open(kwargs['style'], 'r').read()
        except IOError:
            pass

    if kwargs.get('class_fields'):
        kwargs['classes'] = kwargs.pop('class_fields').split(',')

    kwargs.pop('class_fields', None)

    drawing = SVGIS(layers, bounds=bounds, scalar=scalar, use_proj=use_proj,
                    out_crs=out_crs, padding=padding).compose(**kwargs)

    _echo(drawing.tostring(), output)


def _proj(_, output, minx, miny, maxx, maxy, use_proj=None):
    '''Return a transverse mercator projection for the given bounds'''
    prj = projection.generatecrs(minx, miny, maxx, maxy, use_proj)
    _echo(prj + '\n', output)


def main():
    parent = argparse.ArgumentParser(add_help=None)
    parent.add_argument('input', default='/dev/stdin', help="Input svg file")
    parent.add_argument('output', nargs='?', default='/dev/stdout', help="defaults to stdout")

    parser = argparse.ArgumentParser('svgis')
    sp = parser.add_subparsers()

    style = sp.add_parser('style', parents=[parent], help="Add a CSS style to an SVG")
    style.add_argument('-s', '--style', type=str, help="Style string to append to SVG")
    style.add_argument('-r', '--replace', action='store_true', help="Replace the SVG's style")
    style.set_defaults(function=_style)

    scale = sp.add_parser('scale', parents=[parent], help='Scale all coordinates in an SVG by a factor')
    scale.add_argument('-f', '--scale', type=int)
    scale.set_defaults(function=_scale)

    draw = sp.add_parser('draw', help='Draw SVGs from input geodata ')
    draw.add_argument('input', nargs='+', default='/dev/stdin', help="Input geodata layers")
    draw.add_argument('-o', '--output', default='/dev/stdout', help="defaults to stdout")
    draw.add_argument('--bounds', nargs=4, type=float, metavar=('minx', 'miny', 'maxx', 'maxy'),
                      help='In the same coordinate system as the input layers', default=None)

    draw.add_argument('-c', '--style', type=str, metavar='CSS', help="CSS file or string")
    draw.add_argument('-f', '--scale', type=int, default=1,
                      help='Scale for the map (units are divided by this number)')
    draw.add_argument('-p', '--padding', type=int, default=0, required=None,
                      help='Buffer the map bounds (in projection units)')

    draw.add_argument('-x', '--no-viewbox', action='store_false', dest='viewbox',
                      help='Draw SVG without a ViewBox. May improve compatibility.')

    draw.add_argument('--id-field', type=str, dest='id_field', help='Geodata field to use as ID')
    draw.add_argument('--class', type=str, dest='class_fields',
                      help='Geodata fields to use as class (comma-separated)')

    group = draw.add_mutually_exclusive_group()
    group.add_argument('-g', '--epsg', type=str, help='EPSG code to use in projecting output')
    group.add_argument('-j', '--proj4', type=str, help='Proj4 string defining projection use in output')
    group.add_argument('-m', '--projection-method', choices=('utm', 'local'), type=str, dest='use_proj',
                       help=('Projection to use: ',
                             'either the local UTM zone, or a custom Transverse Mercator projection'
                             'centered on the bounding box'))

    draw.set_defaults(function=_draw)

    proj = sp.add_parser('project', help='Get a local Transverse Mercator projection for a bounding box')
    proj.add_argument('minx', type=float, help='west')
    proj.add_argument('miny', type=float, help='south')
    proj.add_argument('maxx', type=float, help='east')
    proj.add_argument('maxy', type=float, help='north')
    proj.add_argument('-m', '--projection-method', dest='use_proj', choices=('utm', 'local'), type=str,)
    proj.set_defaults(function=_proj, input=None, output='/dev/stdout')

    args = parser.parse_args()

    non_keywords = ('function', 'layer', 'output', 'input')
    kwargs = {k: v for k, v in vars(args).items() if k not in non_keywords}

    if args.input in ('-', '/dev/stdin'):
        args.input = sys.stdin

    if args.output in ('-', '/dev/stdout'):
        args.output = sys.stdout

    args.function(args.input, args.output, **kwargs)


if __name__ == '__main__':
    main()



'''Helpers for dealing with projections'''
# -*- coding: utf-8 -*-
import pyproj
import utm
from fiona import transform
import fiona.crs
from pyproj import Proj
from . import scale


def tm_proj4(x0, y0, y1):
    return ('+proj=lcc +lon_0={x0} +lat_1={y1} +lat_2={y0} +lat_0={y1}'
            '+x_0=0 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0'
            '+units=m +no_defs').format(x0=x0, y0=y0, y1=y1)


def tm(x0, y0, y1):
    '''Generate a local transverse mercator projection'''
    proj4 = tm_proj4(x0, y0, y1)
    return pyproj.Proj(proj4)


def utm_proj4(x, y):
    '''Generate the proj4 string for a given (lon, lat) coordinate'''
    try:
        _, _, zonenumber, zoneletter = utm.from_latlon(y, x)
        return zonetoproj4(zonenumber, zoneletter)

    except utm.error.OutOfRangeError as e:
        raise ValueError(e)


def zonetoproj4(zonenumber, zoneletter):
    if zoneletter in 'ZYXWVUTSRQPN':
        hemisphere = 'north'

    elif zoneletter in 'MLKJHGFEDCBA':
        hemisphere = 'south'

    return '+proj=utm +zone={} +{} +datum=WGS84 +units=m +no_defs'.format(zonenumber, hemisphere)


def generatecrs(minx, miny, maxx, maxy, use_proj=None):
    '''Choose a projection, either the local UTM zone or
    create a custom transverse mercator.
    Returns a proj4 string
    '''
    if use_proj == 'utm':
        midx = (minx + maxx) / 2
        midy = (miny + maxy) / 2

        return utm_proj4(midx, midy)

    else:
        # Create a custom TM projection
        x0 = (float(minx) + float(maxx)) // 2

        return tm_proj4(x0, miny, maxy)


def choosecrs(in_crs, bounds, use_proj=None):
    '''Choose a projection. If the layer is projected, use that.
    Otherwise, create use a passed projection or create a custom transverse mercator.
    :in_crs dict A fiona-type proj4 dictionary
    :bounds tuple (minx, miny, maxx, maxy)
    :use_proj string wither 'utm' or 'local'
    :returns dict fiona-type proj4 dictionary
    '''
    if use_proj is None and not Proj(**in_crs).is_latlong():
        # it's projected already, so noop.
        return in_crs

    else:
        return fiona.crs.from_string(generatecrs(*bounds, use_proj=use_proj))


def project_scale(in_crs, out_crs, ring, scalar=None):
    '''Project and apply a scale to a ring'''
    xs, ys = transform.transform(in_crs, out_crs, *zip(*ring))

    projected = zip(xs, ys)

    # then scale
    if scalar:
        return scale.scale(projected, scalar)
    else:
        return projected


def project_mbr(in_crs, out_crs, minx, miny, maxx, maxy):
    minpt, maxpt = zip(*transform.transform(in_crs, out_crs, (minx, maxx), (miny, maxy)))
    return minpt + maxpt



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of svgis.
# https://github.com/fitnr/svgis

# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2015, Neil Freeman <contact@fakeisthenewreal.org>

import unittest
from svgis import convert

class ConvertTestCase(unittest.TestCase):
    def test_rect(self):
        pass

    def test_updatebounds(self):
        bounds1 = (None, 0.1, None, 1.1)
        bounds2 = (0.2, 0.2, 1.2, 1.2)
        bounds3 = (0.1, 0.3, 1.5, 1.1)

        self.assertSequenceEqual(convert.updatebounds(bounds1, bounds2), (0.2, 0.1, 1.2, 1.2))
        self.assertSequenceEqual(convert.updatebounds(bounds3, bounds2), (0.1, 0.2, 1.5, 1.2))



# -*- coding: utf-8 -*-

"""Draw OSM geometries as SVG"""

from __future__ import division, print_function
try:
    from lxml.etree import ElementTree
except ImportError:
    from xml.etree import ElementTree
import svgwrite
from fiona.transform import transform
from . import projection
from .scale import scale
from . import draw as svgisdraw

CRS = {'init': 'epsg:4326'}


def get_root(osmfile):
    if type(osmfile) == str:
        return ElementTree.parse(osmfile).getroot()
    else:
        return osmfile


def bounds(osmfile):
    '''Calculate the bounds of an osm file or element'''
    osm = get_root(osmfile)
    lons, lats = zip(*[(n.get('lon'), n.get('lat')) for n in osm.findall('node')])

    return float(min(lons)), float(min(lats)), float(max(lons)), float(max(lats))


def draw_way(root, way, project, **kwargs):
    """Draw an OSM way as an svgwrite geometry."""
    scalar = kwargs.get('scalar', 1)

    nodes = way.findall('nd')
    cx, cy = [], []

    for ref in nodes:
        node = root.find('node[@id="' + ref.get('ref') + '"]')
        cx.append(float(node.get('lon')))
        cy.append(float(node.get('lat')))

    px, py = project(cx, cy)

    geometry = {
        'coordinates': scale(zip(px, py), scalar)
    }

    tipe = way.find('tag[@key="type"]')

    if tipe and tipe.get('v', None) == 'multipolygon':
        geometry['type'] = 'MultiPolygon'

    elif geometry['coordinates'][0] == geometry['coordinates'][-1]:
        geometry['type'] = 'Polygon'
        geometry['coordinates'] = [geometry['coordinates']]

    else:
        geometry['type'] = 'LineString'

    return svgisdraw.geometry(geometry, **kwargs)


def draw(osmfile, scalar=None, out_crs=None, **kwargs):
    """Draw ways in a OSM file"""
    osm = get_root(osmfile)

    out_crs = make_projection(osm, out_crs)
    project = lambda xs, ys: transform('EPSG:4326', out_crs, xs, ys)

    group = svgwrite.container.Group()

    for way in osm.findall('way'):
        for elem in draw_way(osm, way, project, scalar=scalar, **kwargs):
            group.add(elem)

    return group


def make_projection(osm, out_crs=None):
    if out_crs is None:
        coords = [(float(n.get('lon')), float(n.get('lat'))) for n in osm.findall('node')]
        lons, lats = zip(*coords)

        midx = (max(lons) + min(lons)) / 2
        midy = (max(lats) + min(lats)) / 2

        out_crs = projection.utm_proj4(midx, midy)

    return out_crs



import unittest
from svgis import draw


class DrawTestCase(unittest.TestCase):

    def setUp(self):
        self.lis1 = [[-110.277906, 35.590313], [-110.271087, 35.590375], [-110.26093, 35.590865], [-110.259592, 35.590881], [-110.250972, 35.590888], [-110.250543, 35.590882], [-110.248341, 35.591048], [-110.247998, 35.591065], [-110.230537, 35.59107], [-110.228281, 35.591044], [-110.2257, 35.591041], [-110.277906, 35.590313]]

    def testDrawPoint(self):
        drawn = draw.point((0, 0), r=2)
        self.assertEqual((drawn.attribs['cx'], drawn.attribs['cy']), (0, 0))
        self.assertEqual(drawn.attribs['r'], 2)

        geom = {
            'coordinates': (0, 0),
            'type': 'Point'
        }
        points = draw.points(geom)
        assert len(points) == 1
        self.assertEqual((points[0].attribs['cx'], points[0].attribs['cy']), (0, 0))

        assert draw.geometry(geom).pop(0).attribs['cy'] == 0
        assert draw.geometry(geom).pop(0).attribs['cx'] == 0

    def testDrawLine(self):
        geom = {
            'coordinates': [(0, 0), (1, 1)],
            'type': 'LineString'
        }

        assert draw.linestring(geom['coordinates']).points == geom['coordinates']

        lines = draw.lines(geom)
        assert len(lines) == 1
        assert lines[0].points == geom['coordinates']

    def testDrawMultiLine(self):
        geom = {
            'coordinates': [[(0, 0), (1, 1)], [(3, 2), (5, 1)]],
            'type': 'MultiLineString'
        }
        mls1 = draw.multilinestring(geom['coordinates'])
        mls2 = draw.lines(geom)

        assert mls1.pop(0).points == geom['coordinates'][0]
        assert mls2.pop(1).points == geom['coordinates'][1]
        assert draw.geometry(geom).pop(1).points == geom['coordinates'][1]

        with self.assertRaises(TypeError):
            draw.linestring(geom['coordinates'])

    def testDrawPolygon(self):
        geom = {
            "type": "Polygon",
            "coordinates": [self.lis1]
        }

        drawn = draw.polygon(geom['coordinates'])

        assert len(drawn.points) == len(geom['coordinates'][0])

        for (px, py), (nx, ny) in zip(drawn.points, geom['coordinates'][0]):
            self.assertAlmostEqual(px, nx, 3)
            self.assertAlmostEqual(py, ny, 3)

    def testDrawMultiPolygon(self):
        lis2 = [[-110.005828, 35.498343], [-110.008466, 35.498454], [-110.014261, 35.498655], [-110.019361, 35.499385], [-110.021204, 35.499652], [-110.025195, 35.50001], [-110.028655, 35.50001], [-110.034659, 35.496338], [-110.032935, 35.49411], [-110.005828, 35.498343]]

        geom = {
            "type": "MultiPolygon",
            "coordinates": [[self.lis1], [lis2]]
        }

        drawn = draw.multipolygon(geom['coordinates'])

        assert type(drawn) == list
        assert len(drawn[0].points) == len(geom['coordinates'][0][0])

        for (px, py), (nx, ny) in zip(drawn[1].points, geom['coordinates'][1][0]):
            self.assertAlmostEqual(px, nx, 3)
            self.assertAlmostEqual(py, ny, 3)
